#ifndef REJECTEDMESSAGE_H
#define REJECTEDMESSAGE_H
/**
 * Entry for tracking messages that the server rejected.  We have to
 * make sure that all clients know that this message was rejected to
 * prevent the server from reusing these messages in an attack.
 * @author Brian Demsky
 * @version 1.0
 */
#include "common.h"
#include "Entry.h"

class RejectedMessage : public Entry {
private:
	/* Sequence number */
	int64_t sequencenum;
	/* Machine identifier */
	int64_t machineid;
	/* Oldest sequence number in range */
	int64_t oldseqnum;
	/* Newest sequence number in range */
	int64_t newseqnum;
	/* Is the machine identifier of the relevant slots equal to (or not
	 * equal to) the specified machine identifier. */
	bool equalto;
	/* Set of machines that have not received notification. */
	Hashset<int64_t> *watchset;

public:
	RejectedMessage(Slot *slot, int64_t _sequencenum, int64_t _machineid, int64_t _oldseqnum, int64_t _newseqnum, bool _equalto) : Entry(slot),
		sequencenum(_sequencenum),
		machineid(_machineid),
		oldseqnum(_oldseqnum),
		newseqnum(_newseqnum),
		equalto(_equalto) {
	}

	int64_t getOldSeqNum() { return oldseqnum; }
	int64_t getNewSeqNum() { return newseqnum; }
	bool getEqual() { return equalto; }
	int64_t getMachineID() { return machineid; }
	int64_t getSequenceNumber() { return sequencenum; }
	void setWatchSet(Hashset<int64_t> *_watchset) { watchset = _watchset;  }
	void removeWatcher(int64_t machineid);
	void encode(ByteBuffer *bb);
	int getSize() { return 4 * sizeof(int64_t) + 2 * sizeof(char); }
	char getType() { return TypeRejectedMessage; }
	Entry *getCopy(Slot *s) {
		return new RejectedMessage(s,sequencenum, machineid, oldseqnum, newseqnum, equalto);
	}
};

Entry *RejectedMessage_decode(Slot *slot, ByteBuffer *bb);
#endif
