#ifndef TRANSACTION_H
#define TRANSACTION_H
#include "common.h"
#include "Pair.h"

class Transaction {
private:
	Hashtable<int32_t, TransactionPart *> *parts;
	Hashset<int32_t> *missingParts;
	Vector<int32_t> *partsPendingSend;
	bool fldisComplete;
	bool hasLastPart;
	Hashset<KeyValue *> *keyValueGuardSet;
	Hashset<KeyValue *> *keyValueUpdateSet;
	bool isDead;
	int64_t sequenceNumber;
	int64_t clientLocalSequenceNumber;
	int64_t arbitratorId;
	int64_t machineId;
	Pair<uint64_t, uint64_t> *transactionId;
	int nextPartToSend;
	bool flddidSendAPartToServer;
	TransactionStatus *transactionStatus;
	bool hadServerFailure;
	void decodeTransactionData();

public:
	Transaction();
	void addPartEncode(TransactionPart *newPart);
	void addPartDecode(TransactionPart *newPart);
	void addUpdateKV(KeyValue *kv);
	void addGuardKV(KeyValue *kv);
	int64_t getSequenceNumber();
	void setSequenceNumber(int64_t _sequenceNumber);
	int64_t getClientLocalSequenceNumber();
	Hashtable<int32_t, TransactionPart *> *getParts();
	bool didSendAPartToServer();
	void resetNextPartToSend();
	TransactionPart *getNextPartToSend();
	void setServerFailure();
	bool getServerFailure();
	void resetServerFailure();
	void setTransactionStatus(TransactionStatus *_transactionStatus);
	TransactionStatus *getTransactionStatus();
	void removeSentParts(Vector<int32_t> *sentParts);
	bool didSendAllParts();
	Hashset<KeyValue *> *getKeyValueUpdateSet();
	int getNumberOfParts();
	int64_t getMachineId();
	int64_t getArbitrator();
	bool isComplete();
	Pair<int64_t, int64_t> *getId();
	void setDead();
	TransactionPart *getPart(int32_t index);
	bool evaluateGuard(Hashtable<IoTString *, KeyValue *> *committedKeyValueTable, Hashtable<IoTString *, KeyValue *> *speculatedKeyValueTable, Hashtable<IoTString *, KeyValue *> *pendingTransactionSpeculatedKeyValueTable);
};
#endif
